"""Module for Generic SmartSheet API data models"""

from datetime import datetime
from typing import Any, List, Optional, Union

from pydantic import BaseModel, field_validator


def _parse_datetime_str(value: Any) -> datetime:
    """Adds handling of datetime strings that end with both offset and Z"""
    if isinstance(value, str) and value.endswith("+00:00Z"):
        return datetime.fromisoformat(value.replace("+00:00Z", "+00:00"))
    elif isinstance(value, str):
        return datetime.fromisoformat(value)
    else:
        return value


class SheetColumn(BaseModel):
    """SmartSheet column information"""

    id: int
    index: int
    title: str
    type: str
    validation: bool
    version: int
    width: int
    options: List[str] = []
    primary: bool = False
    hidden: Optional[bool] = None


class SheetRowCell(BaseModel):
    """SmartSheet 'row.cell' information"""

    columnId: int
    displayValue: Optional[str] = None
    value: Optional[Union[str, float]] = None


class SheetRow(BaseModel):
    """SmartSheet row information"""

    cells: List[SheetRowCell]
    createdAt: datetime
    expanded: bool
    id: int
    modifiedAt: datetime
    rowNumber: int
    siblingId: Optional[int] = None

    validate_fields = field_validator(
        "createdAt", "modifiedAt", mode="before"
    )(_parse_datetime_str)


class SheetFields(BaseModel):
    """SmartSheet information"""

    columns: List[SheetColumn]
    accessLevel: str
    createdAt: datetime
    dependenciesEnabled: bool
    effectiveAttachmentOptions: List[str]
    ganttEnabled: bool
    hasSummaryFields: bool
    id: int
    modifiedAt: datetime
    name: str
    permalink: str
    readOnly: bool
    resourceManagementEnabled: bool
    rows: List[SheetRow]
    totalRowCount: int
    userPermissions: dict
    userSettings: dict
    version: int
    workspace: dict = {}

    validate_fields = field_validator(
        "createdAt", "modifiedAt", mode="before"
    )(_parse_datetime_str)



"""Module to instantiate a client to connect to Smartsheet and provide helpful
methods to retrieve data."""

from functools import lru_cache
from typing import Any, Dict, List, Optional, Type, TypeVar

from pydantic import BaseModel, Field, SecretStr, ValidationError
from pydantic_settings import BaseSettings, SettingsConfigDict
from smartsheet import Smartsheet

from aind_smartsheet_api.models import SheetFields, SheetRow

T = TypeVar("T", bound=BaseModel)


class SmartsheetSettings(BaseSettings):
    """Configuration class. Mostly a wrapper around smartsheet.Smartsheet
    class constructor arguments."""

    access_token: SecretStr = Field(
        ..., description="API token can be created in Smartsheet UI"
    )
    user_agent: Optional[str] = Field(
        default=None,
        description=(
            "The user agent to use when making requests. "
            "Helps identify requests."
        ),
    )
    max_connections: int = Field(
        default=8, description="Maximum connection pool size."
    )
    model_config = SettingsConfigDict(env_prefix="SMARTSHEET_", extra="forbid")


class SmartsheetClient:
    """Main client to connect to a Smartsheet sheet. Requires an API token
    and the sheet id."""

    def __init__(self, smartsheet_settings: SmartsheetSettings):
        """
        Class constructor
        Parameters
        ----------
        smartsheet_settings : SmartsheetSettings
        """
        self.smartsheet_settings = smartsheet_settings
        self.smartsheet_client = Smartsheet(
            user_agent=self.smartsheet_settings.user_agent,
            max_connections=self.smartsheet_settings.max_connections,
            access_token=(
                self.smartsheet_settings.access_token.get_secret_value()
            ),
        )

    @lru_cache(maxsize=1)
    def get_raw_sheet(self, sheet_id: int) -> str:
        """
        Retrieve the sheet as a json dictionary..
        Parameters
        ----------
        sheet_id : int

        Returns
        -------
        str
          Raises an error if there is an issue retrieving the sheet.

        """
        smartsheet_response = self.smartsheet_client.Sheets.get_sheet(sheet_id)
        return smartsheet_response.to_json()

    def get_sheet_fields(self, sheet_id: int) -> SheetFields:
        """

        Parameters
        ----------
        sheet_id : int

        Returns
        -------
        SheetFields

        """
        raw_sheet = self.get_raw_sheet(sheet_id=sheet_id)
        return SheetFields.model_validate_json(raw_sheet)

    @staticmethod
    def _column_id_map(sheet_fields: SheetFields) -> Dict[int, str]:
        """
        SmartSheet uses integer ids for the columns. We need a way to
        map the column titles to the ids so we can retrieve information using
        just the titles.
        Parameters
        ----------
        sheet_fields : SheetFields

        Returns
        -------
        Dict[int, str]

        """
        return {c.id: c.title for c in sheet_fields.columns}

    @staticmethod
    def _map_row_to_dict(
        row: SheetRow, column_id_map: Dict[int, str]
    ) -> Dict[str, Any]:
        """
        Maps a row into a dictionary that maps the column names to their values
        Parameters
        ----------
        row : SheetRow
          A SheetRow that will be parsed. This a list of cells with a columnId
          and a cell value.
        column_id_map: Dict[int, str]
          Map of column ids into the column names

        Returns
        -------
        Dict[str, Any]
          The list of row cells is converted to a dictionary.

        """
        output_dict = {}
        for cell in row.cells:
            column_id = cell.columnId
            column_name = column_id_map[column_id]
            output_dict[column_name] = cell.value
        return output_dict

    def get_parsed_sheet(self, sheet_id: int) -> List[Dict[str, Any]]:
        """
        Parse raw sheet json into basic dictionary of {"name": "value"}
        Parameters
        ----------
        sheet_id : int

        Returns
        -------
        Dict[str, Any]

        """

        sheet_fields = self.get_sheet_fields(sheet_id=sheet_id)
        column_id_map = self._column_id_map(sheet_fields=sheet_fields)
        parsed_rows = list()
        for row in sheet_fields.rows:
            row_dict = self._map_row_to_dict(row, column_id_map=column_id_map)
            parsed_rows.append(row_dict)
        return parsed_rows

    def get_parsed_sheet_model(
        self, sheet_id: int, model: Type[T], validate: bool = True
    ) -> List[Type[T]]:
        """
        Parse raw sheet json into basic dictionary of {"name": "value"}
        Parameters
        ----------
        sheet_id : int
        model : T
          BaseModel type
        validate : bool
          If set to True, will raise errors if pydantic validation fails.
          If set to False, will use model_construct in places where validation
          fails. Default is True.

        Returns
        -------
        List[Type[T]]

        """

        sheet_fields = self.get_sheet_fields(sheet_id=sheet_id)
        column_id_map = self._column_id_map(sheet_fields=sheet_fields)
        parsed_rows = list()
        for row in sheet_fields.rows:
            row_dict = self._map_row_to_dict(row, column_id_map=column_id_map)
            try:
                row_as_model = model.model_validate(row_dict)
                parsed_rows.append(row_as_model)

            except ValidationError as e:
                if validate:
                    raise e
                else:
                    row_as_model = model.model_construct(**row_dict)
                    parsed_rows.append(row_as_model)
        return parsed_rows



"""Tests classes in models module"""

import json
import os
import unittest
from pathlib import Path
from typing import Optional
from unittest.mock import MagicMock, call, patch

from pydantic import BaseModel, ConfigDict, Field, ValidationError

from aind_smartsheet_api.client import SmartsheetClient, SmartsheetSettings

RESOURCE_DIR = Path(os.path.dirname(os.path.realpath(__file__))) / "resources"
EXAMPLE_SHEET = RESOURCE_DIR / "example_sheet.json"


class TestSmartsheetSettings(unittest.TestCase):
    """Tests configs for SmartsheetSettings."""

    @patch.dict(
        os.environ,
        {
            "SMARTSHEET_ACCESS_TOKEN": "abc-123",
            "SMARTSHEET_USER_AGENT": "secret_agent",
        },
        clear=True,
    )
    def test_env_var(self):
        """Tests class can be set via env vars."""
        smart_sheet_settings = SmartsheetSettings()
        self.assertEqual(
            "abc-123", smart_sheet_settings.access_token.get_secret_value()
        )
        self.assertEqual("secret_agent", smart_sheet_settings.user_agent)


class TestSmartsheetClient(unittest.TestCase):
    """Tests methods in SmartsheetClient class"""

    class MockSheetModel1(BaseModel):
        """Mocked class to hold sheet model with all valid fields for mocked
        sheet contents"""

        project_name: Optional[str] = Field(None, alias="Project Name")
        project_code: str = Field(..., alias="Project Code")
        funding_institution: str = Field(..., alias="Funding Institution")
        grant_number: Optional[str] = Field(None, alias="Grant Number")
        investigators: str = Field(..., alias="Investigators")
        model_config = ConfigDict(populate_by_name=True)

    class MockSheetModel2(BaseModel):
        """Mocked class to hold sheet model with some invalid fields for
        mocked sheet contents"""

        project_name: str = Field(..., alias="Project Name")
        project_code: str = Field(..., alias="Project Code")
        funding_institution: str = Field(..., alias="Funding Institution")
        # Some grant numbers are None, so this will be invalid
        grant_number: str = Field(..., alias="Grant Number")
        investigators: str = Field(..., alias="Investigators")
        model_config = ConfigDict(populate_by_name=True)

    @classmethod
    def setUpClass(cls) -> None:
        """Read in json files for mock responses"""
        with open(EXAMPLE_SHEET, "r") as f:
            contents = json.load(f)
        cls.example_sheet: str = json.dumps(contents)
        default_settings = SmartsheetSettings(access_token="abc-123")
        cls.default_client = SmartsheetClient(
            smartsheet_settings=default_settings
        )

    @patch("smartsheet.sheets.Sheets.get_sheet")
    def test_get_sheet_cache(self, mock_get_sheet: MagicMock):
        """Tests get_sheet pulls from cache if called twice"""
        mock_get_sheet.return_value.to_json.return_value = self.example_sheet
        default_settings = SmartsheetSettings(access_token="abc-123")
        client = SmartsheetClient(smartsheet_settings=default_settings)

        response1 = client.get_raw_sheet(sheet_id=2014892103339908)
        response2 = client.get_raw_sheet(sheet_id=2014892103339908)
        self.assertEqual(response1, response2)
        mock_get_sheet.assert_has_calls(
            [call(2014892103339908), call().to_json()]
        )

    @patch("smartsheet.sheets.Sheets.get_sheet")
    def test_get_sheet_model(self, mock_get_sheet: MagicMock):
        """Tests get_sheet_model method"""
        mock_get_sheet.return_value.to_json.return_value = self.example_sheet
        sheet_fields = self.default_client.get_sheet_fields(
            sheet_id=2014892103339908
        )
        self.assertEqual(5, len(sheet_fields.columns))
        self.assertEqual(3, len(sheet_fields.rows))

    @patch("smartsheet.sheets.Sheets.get_sheet")
    def test_column_id_map(self, mock_get_sheet: MagicMock):
        """Tests _column_id_map method"""
        mock_get_sheet.return_value.to_json.return_value = self.example_sheet
        sheet_fields = self.default_client.get_sheet_fields(
            sheet_id=2014892103339908
        )
        col_id_map = self.default_client._column_id_map(
            sheet_fields=sheet_fields
        )
        expected_col_id_map = {
            3981351074090884: "Project Name",
            1729551260405636: "Project Code",
            2990791841501060: "Funding Institution",
            3446515788894084: "Grant Number",
            4825776004222852: "Investigators",
        }
        self.assertEqual(expected_col_id_map, col_id_map)

    @patch("smartsheet.sheets.Sheets.get_sheet")
    def test_map_row_to_dict(self, mock_get_sheet: MagicMock):
        """Tests _map_row_to_dict method"""

        mock_get_sheet.return_value.to_json.return_value = self.example_sheet
        sheet_fields = self.default_client.get_sheet_fields(
            sheet_id=2014892103339908
        )
        col_id_map = self.default_client._column_id_map(
            sheet_fields=sheet_fields
        )
        first_row = sheet_fields.rows[0]
        row_as_dict = self.default_client._map_row_to_dict(
            row=first_row, column_id_map=col_id_map
        )
        expected_row_as_dict = {
            "Project Name": "AIND Scientific Activities",
            "Project Code": "122-01-001-10",
            "Funding Institution": "Allen Institute",
            "Grant Number": None,
            "Investigators": "person.two@acme.org, J Smith, John Doe II",
        }
        self.assertEqual(expected_row_as_dict, row_as_dict)

    @patch("smartsheet.sheets.Sheets.get_sheet")
    def test_get_parsed_sheet(self, mock_get_sheet: MagicMock):
        """Tests get_parsed_sheet method"""

        mock_get_sheet.return_value.to_json.return_value = self.example_sheet
        parsed_sheet = self.default_client.get_parsed_sheet(
            sheet_id=2014892103339908
        )
        expected_output = [
            {
                "Project Name": "AIND Scientific Activities",
                "Project Code": "122-01-001-10",
                "Funding Institution": "Allen Institute",
                "Grant Number": None,
                "Investigators": "person.two@acme.org, J Smith, John Doe II",
            },
            {
                "Project Name": None,
                "Project Code": "122-01-001-10",
                "Funding Institution": "Allen Institute",
                "Grant Number": None,
                "Investigators": "John Doe, person.one@acme.org",
            },
            {
                "Project Name": "v1omFISH",
                "Project Code": "121-01-010-10",
                "Funding Institution": "Allen Institute",
                "Grant Number": None,
                "Investigators": "person.one@acme.org, Jane Doe",
            },
        ]

        self.assertEqual(expected_output, parsed_sheet)

    @patch("smartsheet.sheets.Sheets.get_sheet")
    def test_get_parsed_sheet_model_case_1(self, mock_get_sheet: MagicMock):
        """Tests get_parsed_sheet_model method with validate set to True and
        no validation errors"""

        mock_get_sheet.return_value.to_json.return_value = self.example_sheet
        parsed_sheet = self.default_client.get_parsed_sheet_model(
            sheet_id=2014892103339908,
            model=self.MockSheetModel1,
            validate=True,
        )
        expected_output = [
            self.MockSheetModel1(
                project_name="AIND Scientific Activities",
                project_code="122-01-001-10",
                funding_institution="Allen Institute",
                grant_number=None,
                investigators="person.two@acme.org, J Smith, John Doe II",
            ),
            self.MockSheetModel1(
                project_name=None,
                project_code="122-01-001-10",
                funding_institution="Allen Institute",
                grant_number=None,
                investigators="John Doe, person.one@acme.org",
            ),
            self.MockSheetModel1(
                project_name="v1omFISH",
                project_code="121-01-010-10",
                funding_institution="Allen Institute",
                grant_number=None,
                investigators="person.one@acme.org, Jane Doe",
            ),
        ]
        self.assertEqual(expected_output, parsed_sheet)

    @patch("smartsheet.sheets.Sheets.get_sheet")
    def test_get_parsed_sheet_model_case_2(self, mock_get_sheet: MagicMock):
        """Tests get_parsed_sheet_model method with validate set to True and
        validation errors"""

        mock_get_sheet.return_value.to_json.return_value = self.example_sheet
        with self.assertRaises(ValidationError):
            self.default_client.get_parsed_sheet_model(
                sheet_id=2014892103339908,
                model=self.MockSheetModel2,
                validate=True,
            )

    @patch("smartsheet.sheets.Sheets.get_sheet")
    def test_get_parsed_sheet_model_case_3(self, mock_get_sheet: MagicMock):
        """Tests get_parsed_sheet_model method with validate set to False and
        validation errors"""

        mock_get_sheet.return_value.to_json.return_value = self.example_sheet
        parsed_sheet = self.default_client.get_parsed_sheet_model(
            sheet_id=2014892103339908,
            model=self.MockSheetModel2,
            validate=False,
        )
        expected_output = [
            self.MockSheetModel2.model_construct(
                project_name="AIND Scientific Activities",
                project_code="122-01-001-10",
                funding_institution="Allen Institute",
                grant_number=None,
                investigators="person.two@acme.org, J Smith, John Doe II",
            ),
            self.MockSheetModel2.model_construct(
                project_name=None,
                project_code="122-01-001-10",
                funding_institution="Allen Institute",
                grant_number=None,
                investigators="John Doe, person.one@acme.org",
            ),
            self.MockSheetModel2.model_construct(
                project_name="v1omFISH",
                project_code="121-01-010-10",
                funding_institution="Allen Institute",
                grant_number=None,
                investigators="person.one@acme.org, Jane Doe",
            ),
        ]
        self.assertEqual(expected_output, parsed_sheet)


if __name__ == "__main__":
    unittest.main()



"""Tests classes in models module"""

import json
import os
import unittest
from datetime import datetime, timezone
from pathlib import Path

from aind_smartsheet_api.models import SheetFields, SheetRow, SheetRowCell

RESOURCE_DIR = Path(os.path.dirname(os.path.realpath(__file__))) / "resources"
EXAMPLE_SHEET = RESOURCE_DIR / "example_sheet.json"


class TestParsing(unittest.TestCase):
    """Tests validators for SheetRow and SheetFields class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Read example json files"""
        with open(EXAMPLE_SHEET, "r") as f:
            contents = json.load(f)
        cls.example_sheet: str = json.dumps(contents)

    def test_parse_datetime_str(self):
        """Tests _parse_datetime_str validator function."""
        example_row = SheetRow(
            cells=[
                SheetRowCell(
                    columnId=1729551260405636,
                    value="122-01-001-10",
                    displayValue="122-01-001-10",
                )
            ],
            createdAt="2023-08-07T23:39:14+00:00Z",
            expanded=False,
            id=123456,
            modifiedAt="2023-12-20T18:32:10+00:00",
            rowNumber=1,
        )
        expected_validated_row = SheetRow(
            cells=[
                SheetRowCell(
                    columnId=1729551260405636,
                    value="122-01-001-10",
                    displayValue="122-01-001-10",
                )
            ],
            createdAt=datetime(2023, 8, 7, 23, 39, 14, tzinfo=timezone.utc),
            expanded=False,
            id=123456,
            modifiedAt=datetime(2023, 12, 20, 18, 32, 10, tzinfo=timezone.utc),
            rowNumber=1,
        )
        self.assertEqual(expected_validated_row, example_row)

    def test_sheet_fields(self):
        """Tests SheetFields parses json file into generic model"""
        json_contents = self.example_sheet
        sheet_fields = SheetFields.model_validate_json(json_contents)
        self.assertEqual(5, len(sheet_fields.columns))
        self.assertEqual(3, len(sheet_fields.rows))


if __name__ == "__main__":
    unittest.main()
